﻿using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Kinect; // KINECT

namespace XNA4KinectHandler
{
    public enum Body
    {
        Head_point,
        Neck_point,
        LeftHip_point,
        LeftFoot_point,
        LeftElbow_point,
        LeftKnee_point,
        LeftShoulder_point,
        LeftHand_point,
        RightHip_point,
        RightFoot_point,
        RightElbow_point,
        RightKnee_point,
        RightShoulder_point,
        RightHand_point
    }

    public class KinectHandler : DrawableGameComponent
    {
        static Random rand = new Random();

        KinectSensor kinectSensor;                  // kinect sensor
        SkeletonFrame skeletonFrame;                // skeleton
        //public Effect kinectEffect; // kinect FX
        string connectedStatus = "Not connected";// kinect status
        //SpriteFont font;                            // kinect status

        // left hand
        public Vector3 CurrentLeftHand;
        public Vector3 PrevLeftHand; 
        private Vector2 leftHandMoved;
        public Vector2 LeftHandMoved
        {
            get { return leftHandMoved; }
        }
        // right hand
        public Vector3 CurrentRightHand;
        public Vector3 PrevRightHand;
        private Vector2 rightHandMoved;
        public Vector2 RightHandMoved
        {
            get { return rightHandMoved; }
        }
        // left foot
        public Vector3 CurrentLeftFoot;
        // right foot
        public Vector3 CurrentRightFoot;

        public Vector3[] fullbodySkeletonPoints; // 3D fullbody points
        
        public int[] fullbodySpriteIndices = new int[5]; // 2D fullbody sprits
        public List<SpriteImageNode> fullbodySpriteImages = new List<SpriteImageNode>(); // 각 텍스쳐 이미지 객체
        public bool fullbodySpriteDraw
        {
            get;
            set;
        }
        public Vector2 ScreenCenter
        {
            get;
            set;
        }
        public bool fullbodyTrackingFlag;
        float fullbodyTrackingTimer = 0;

        SkeletonTrackingBuffer[] tb = null;
        Vector3 setTrackingBuffer(Vector3 value, int i)
        {
            return tb[i].getAveg(value);
        }

        public KinectHandler(Game game) : base(game)
        {
            fullbodySkeletonPoints = new Vector3[14];
            // Player 위치정보 초기화
            for (int i = 0; i < fullbodySkeletonPoints.Length; i++)
                fullbodySkeletonPoints[i] = new Vector3(0.0f, 0.0f, -10000);
            tb = new SkeletonTrackingBuffer[14];
            for (int i = 0; i < tb.Length; i++)
            {
                tb[i] = new SkeletonTrackingBuffer();
            }
            
            fullbodyTrackingFlag = false;
            fullbodySpriteDraw = true;
        }

        private void DiscoverKinectSensor()
        {
            foreach (KinectSensor sensor in KinectSensor.KinectSensors)
            {
                if (sensor.Status == KinectStatus.Connected)
                {
                    // Found one, set our sensor to this
                    kinectSensor = sensor;
                    break;
                }
            }

            if (this.kinectSensor == null)
            {
                connectedStatus = "Found none Kinect Sensors connected to USB";
                return;
            }

            // You can use the kinectSensor.Status to check for status
            // and give the user some kind of feedback
            switch (kinectSensor.Status)
            {
                case KinectStatus.Connected:
                {
                    connectedStatus = "Status: Connected";
                    break;
                }
                case KinectStatus.Disconnected:
                {
                    connectedStatus = "Status: Disconnected";
                    break;
                }
                case KinectStatus.NotPowered:
                {
                    connectedStatus = "Status: Connect the power";
                    break;
                }
                default:
                {
                    connectedStatus = "Status: Error";
                    break;
                }
            }

            // Init the found and connected device
            if (kinectSensor.Status == KinectStatus.Connected)
            {
                InitializeKinect();
            }
        }

        private bool InitializeKinect()
        {
            // Skeleton Stream
            kinectSensor.SkeletonStream.Enable(new TransformSmoothParameters()
            {
                Smoothing = 0.5f,
                Correction = 0.5f,
                Prediction = 0.5f,
                JitterRadius = 0.05f,
                MaxDeviationRadius = 0.04f
            });
            kinectSensor.SkeletonFrameReady += new EventHandler<SkeletonFrameReadyEventArgs>(kinectSensor_SkeletonFrameReady);

            try
            {
                kinectSensor.Start();
            }
            catch
            {
                connectedStatus = "Unable to start the Kinect Sensor";
                return false;
            }
            return true;
        }

        private void KinectSensors_StatusChanged(object sender, StatusChangedEventArgs e)
        {
            if (this.kinectSensor == e.Sensor)
            {
                if (e.Status == KinectStatus.Disconnected || e.Status == KinectStatus.NotPowered)
                {
                    this.kinectSensor = null;
                    this.DiscoverKinectSensor();
                }
            }
        }

        public override void Initialize()
        {
            KinectSensor.KinectSensors.StatusChanged += new EventHandler<StatusChangedEventArgs>(KinectSensors_StatusChanged);
            DiscoverKinectSensor();

            base.Initialize();
        }

        // kinect 3D skeleton tracking
        private void kinectSensor_SkeletonFrameReady(object sender, SkeletonFrameReadyEventArgs e)
        {
            using (skeletonFrame = e.OpenSkeletonFrame())
            {
                if (skeletonFrame != null)
                {
                    Skeleton[] skeletonData = new Skeleton[skeletonFrame.SkeletonArrayLength];
                    skeletonFrame.CopySkeletonDataTo(skeletonData);
                    Skeleton playerSkeleton = (from s in skeletonData where s.TrackingState == SkeletonTrackingState.Tracked select s).FirstOrDefault();

                    if (playerSkeleton != null)
                    {
                        Joint head = playerSkeleton.Joints[JointType.Head];
                        fullbodySkeletonPoints[(int)Body.Head_point] = setTrackingBuffer(new Vector3(head.Position.X, head.Position.Y, head.Position.Z * 1.25f), 0);
                        Joint neck = playerSkeleton.Joints[JointType.ShoulderCenter];
                        fullbodySkeletonPoints[(int)Body.Neck_point] = setTrackingBuffer(new Vector3(neck.Position.X, neck.Position.Y - 0.14f, neck.Position.Z * 1.25f), 1);

                        Joint leftHip = playerSkeleton.Joints[JointType.HipLeft];
                        fullbodySkeletonPoints[(int)Body.LeftHip_point] = setTrackingBuffer(new Vector3(leftHip.Position.X, leftHip.Position.Y - 0.12f, leftHip.Position.Z * 1.25f), 2);
                        Joint leftFoot = playerSkeleton.Joints[JointType.FootLeft];
                        fullbodySkeletonPoints[(int)Body.LeftFoot_point] = setTrackingBuffer(new Vector3(leftFoot.Position.X, leftFoot.Position.Y, leftFoot.Position.Z * 1.25f), 3);
                        Joint leftElbow = playerSkeleton.Joints[JointType.ElbowLeft];
                        fullbodySkeletonPoints[(int)Body.LeftElbow_point] = setTrackingBuffer(new Vector3(leftElbow.Position.X, leftElbow.Position.Y, leftElbow.Position.Z * 1.25f), 4);
                        Joint leftKnee = playerSkeleton.Joints[JointType.KneeLeft];
                        fullbodySkeletonPoints[(int)Body.LeftKnee_point] = setTrackingBuffer(new Vector3(leftKnee.Position.X, leftKnee.Position.Y, leftKnee.Position.Z * 1.25f), 5);
                        Joint leftShoulder = playerSkeleton.Joints[JointType.ShoulderLeft];
                        fullbodySkeletonPoints[(int)Body.LeftShoulder_point] = setTrackingBuffer(new Vector3(leftShoulder.Position.X, leftShoulder.Position.Y, leftShoulder.Position.Z * 1.25f), 6);
                        Joint leftHand = playerSkeleton.Joints[JointType.HandLeft];
                        fullbodySkeletonPoints[(int)Body.LeftHand_point] = setTrackingBuffer(new Vector3(leftHand.Position.X, leftHand.Position.Y, leftHand.Position.Z * 1.25f), 7);

                        Joint rightHip = playerSkeleton.Joints[JointType.HipRight];
                        fullbodySkeletonPoints[(int)Body.RightHip_point] = setTrackingBuffer(new Vector3(rightHip.Position.X, rightHip.Position.Y - 0.12f, rightHip.Position.Z * 1.25f), 8);
                        Joint rightFoot = playerSkeleton.Joints[JointType.FootRight];
                        fullbodySkeletonPoints[(int)Body.RightFoot_point] = setTrackingBuffer(new Vector3(rightFoot.Position.X, rightFoot.Position.Y, rightFoot.Position.Z * 1.25f), 9);
                        Joint rightElbow = playerSkeleton.Joints[JointType.ElbowRight];
                        fullbodySkeletonPoints[(int)Body.RightElbow_point] = setTrackingBuffer(new Vector3(rightElbow.Position.X, rightElbow.Position.Y, rightElbow.Position.Z * 1.25f), 10);
                        Joint rightKnee = playerSkeleton.Joints[JointType.KneeRight];
                        fullbodySkeletonPoints[(int)Body.RightKnee_point] = setTrackingBuffer(new Vector3(rightKnee.Position.X, rightKnee.Position.Y, rightKnee.Position.Z * 1.25f), 11);
                        Joint rightShoulder = playerSkeleton.Joints[JointType.ShoulderRight];
                        fullbodySkeletonPoints[(int)Body.RightShoulder_point] = setTrackingBuffer(new Vector3(rightShoulder.Position.X, rightShoulder.Position.Y, rightShoulder.Position.Z * 1.25f), 12);
                        Joint rightHand = playerSkeleton.Joints[JointType.HandRight];
                        fullbodySkeletonPoints[(int)Body.RightHand_point] = setTrackingBuffer(new Vector3(rightHand.Position.X, rightHand.Position.Y, rightHand.Position.Z * 1.25f), 13);
                        System.Diagnostics.Trace.WriteLine("skeleton: rightHand=" + rightHand.Position.X + " " + rightHand.Position.Y + " " + rightHand.Position.Z);

                        fullbodyTrackingFlag = true;
                        return;
                    }
                    else
                    {
                        fullbodyTrackingFlag = false;
                    }
                }
            }
        }

        protected override void LoadContent()
        {
            //kinectEffect = Game.Content.Load<Effect>("Effect/KinectEffect");

            //트래킹 화면 빛 구체
            SpriteBatch spriteBatch = new SpriteBatch(GraphicsDevice);
            fullbodySpriteImages.Add(new SpriteImageNode(spriteBatch, Game.Content.Load<Texture2D>("Texture\\twinkle\\twinkle1")));  // 1~10 빛무리
            fullbodySpriteImages.Add(new SpriteImageNode(spriteBatch, Game.Content.Load<Texture2D>("Texture\\twinkle\\twinkle2")));
            fullbodySpriteImages.Add(new SpriteImageNode(spriteBatch, Game.Content.Load<Texture2D>("Texture\\twinkle\\twinkle3")));
            fullbodySpriteImages.Add(new SpriteImageNode(spriteBatch, Game.Content.Load<Texture2D>("Texture\\twinkle\\twinkle4")));
            fullbodySpriteImages.Add(new SpriteImageNode(spriteBatch, Game.Content.Load<Texture2D>("Texture\\twinkle\\twinkle5")));
            fullbodySpriteImages.Add(new SpriteImageNode(spriteBatch, Game.Content.Load<Texture2D>("Texture\\twinkle\\twinkle6")));
            fullbodySpriteImages.Add(new SpriteImageNode(spriteBatch, Game.Content.Load<Texture2D>("Texture\\twinkle\\twinkle7")));
            fullbodySpriteImages.Add(new SpriteImageNode(spriteBatch, Game.Content.Load<Texture2D>("Texture\\twinkle\\twinkle8")));
            fullbodySpriteImages.Add(new SpriteImageNode(spriteBatch, Game.Content.Load<Texture2D>("Texture\\twinkle\\twinkle9")));
            fullbodySpriteImages.Add(new SpriteImageNode(spriteBatch, Game.Content.Load<Texture2D>("Texture\\twinkle\\twinkle10")));
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
            this.kinectSensor.Stop();
            this.kinectSensor.Dispose();
        }

        public override void Update(GameTime gameTime)
        {
            if (!fullbodyTrackingFlag) //트래킹중이 아닐때 트래킹모드용 빛텍스쳐를 랜덤으로 설정
            {
                for (int i = 0; i < 5; i++)
                {
                    fullbodySpriteIndices[i] = rand.Next(1, 10);
                    fullbodyTrackingTimer = 0;
                }
            }
            else
            {
                fullbodyTrackingTimer += (float)gameTime.ElapsedGameTime.TotalMilliseconds;

                // left hand
                PrevLeftHand = CurrentLeftHand;
                CurrentLeftHand = fullbodySkeletonPoints[(int)Body.LeftHand_point];
                leftHandMoved = new Vector2(PrevLeftHand.X - CurrentLeftHand.X, PrevLeftHand.Y - CurrentLeftHand.Y);
                System.Diagnostics.Trace.WriteLine("leftHandMoved=" + leftHandMoved);
                // right hand
                PrevRightHand = CurrentRightHand;
                CurrentRightHand = fullbodySkeletonPoints[(int)Body.RightHand_point];
                rightHandMoved = new Vector2(PrevRightHand.X - CurrentRightHand.X, PrevRightHand.Y - CurrentRightHand.Y);
                System.Diagnostics.Trace.WriteLine("rightHandMoved=" + rightHandMoved);
                // left foot
                CurrentLeftFoot = fullbodySkeletonPoints[(int)Body.LeftFoot_point];
                // right foot
                CurrentRightFoot = fullbodySkeletonPoints[(int)Body.RightFoot_point];
            }
            base.Update(gameTime);
        }

        public override void Draw(GameTime gameTime)
        {
            if (fullbodyTrackingFlag && fullbodySpriteDraw)
            {
                fullbodySpriteImages[fullbodySpriteIndices[0]].Draw(new Vector2(ScreenCenter.X, ScreenCenter.Y - 160));
                fullbodySpriteImages[fullbodySpriteIndices[1]].Draw(new Vector2(ScreenCenter.X - 120 + (CurrentLeftHand.X * 200), ScreenCenter.Y));
                fullbodySpriteImages[fullbodySpriteIndices[2]].Draw(new Vector2(ScreenCenter.X + 120 + (CurrentRightHand.X * 200), ScreenCenter.Y));
                fullbodySpriteImages[fullbodySpriteIndices[3]].Draw(new Vector2(ScreenCenter.X - 80 + (CurrentLeftFoot.X * 200), ScreenCenter.Y + 200));
                fullbodySpriteImages[fullbodySpriteIndices[4]].Draw(new Vector2(ScreenCenter.X + 80 + (CurrentRightFoot.X * 200), ScreenCenter.Y + 200));
            }
            base.Draw(gameTime);
        }
    }
}
